(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright (c) 2013-2014 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uarray_ng.ml,v 1.3 2014/05/06 12:56:17 fclement Exp $ *)

module Nat = Unatural;;

let a_qual = Printf.sprintf "array_%s";;

let a_invalid_arg = Upervasives.invalid_arg_qual a_qual;;

type index = Upervasives.index;;
type offset = index;;
type length = Upervasives.length;;

let array_get a i = Array.get a (Nat.to_int i);;
let array_unsafe_get a i = Array.unsafe_get a (Nat.to_int i);;
let array_set a i = Array.set a (Nat.to_int i);;
let array_unsafe_set a i = Array.unsafe_set a (Nat.to_int i);;

let array_get_last a =
  Array.unsafe_get a (pred (Nat.to_int (Uarray.length a)))
;;

let array_sub a ofs len = Array.sub a (Nat.to_int ofs) (Nat.to_int len);;

let array_check_subarray fun_name a ofs len =
  if Nat.add ofs len > Uarray.length a then
    a_invalid_arg fun_name "wrong subarray specification"
;;

let array_gen_cum_sub_fold_left_i fun_name f accu a ofs len =
  array_check_subarray fun_name a ofs len;
  let r = ref accu in
  let res = Uarray.make (Nat.succ len) accu in
  let rec loop i imax =
    if i >= imax then () else
    let i' = Nat.add ofs i in
    let i = Nat.succ i in
    r := f !r i' (array_unsafe_get a i');
    array_unsafe_set res i !r;
    loop i imax in
  loop Nat._0_ len;
  res
;;

let array_cum_sub_fold_left_i f =
  array_gen_cum_sub_fold_left_i "cum_sub_fold_left_i" f
;;

let array_cum_sub_fold_left f =
  array_gen_cum_sub_fold_left_i "cum_sub_fold_left"
    (fun accu _i x -> f accu x)
;;

let array_cum_fold_left_i f accu a =
  array_gen_cum_sub_fold_left_i "cum_fold_left_i"
    f accu a Nat._0_ (Uarray.length a)
;;

let array_cum_fold_left f accu a =
  array_gen_cum_sub_fold_left_i "cum_fold_left"
    (fun accu _i x -> f accu x) accu a Nat._0_ (Uarray.length a)
;;

let array_for_all_i p a =
  let len = Uarray.length a in
  let rec loop i = i = len || p i (array_get a i) && loop (Nat.succ i) in
  loop Nat._0_
;;

(* Operations on arrays of arrays. *)

let aa_qual = Printf.sprintf "array_array_%s";;

let aa_invalid_arg = Upervasives.invalid_arg_qual aa_qual;;

type 'a array_array_t = 'a array array;;

let array_array_map f = Array.map (Array.map f);;
let array_array_map2 f = Uarray.map2 (Uarray.map2 f);;

let array_array_pack aa =
  let lens = Array.map Uarray.length aa in
  let a = Array.concat (Array.to_list aa) in
  (lens, a)
;;

let there_is_zero = Uarray.exists (fun n -> n = Nat._0_);;

let incompatible_lengths len a = Uarray.length a <> len;;

let array_array_unpack lens a =
  if there_is_zero lens then
    aa_invalid_arg "unpack" "zero length" else
  let offsets = array_cum_fold_left Nat.add Nat._0_ lens in
  if incompatible_lengths (array_get_last offsets) a then
    aa_invalid_arg "unpack" "incompatible lengths" else
  let offsets = array_sub offsets Nat._0_ (Uarray.length lens) in
  Uarray.map2 (array_sub a) offsets lens
;;

(* More matrix operations. *)

type line_index = index;;
type column_index = index;;

let matrix_get_line m = array_get (Umatrix.to_array_array m);;

let matrix_of_array ni nj a = Umatrix.of_array a ni nj;;

let matrix_map3 f m1 m2 m3 =
  Umatrix.of_array_array
    (Uarray.map3 (Uarray.map3 f)
       (Umatrix.to_array_array m1)
       (Umatrix.to_array_array m2)
       (Umatrix.to_array_array m3))
;;
