(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright (c) 2013-2014 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: display.ml,v 1.30 2014/07/08 15:18:27 fclement Exp $ *)

module Nat = Unatural;;

module Set = Umatrix_nonempty_set;;
module Mapping = Umatrix_graph;;
module Structural_mapping = Umatrix_structural_mapping;;

module Config = Configuration;;

let print_report ident printer ppf x =
  Config.print_report (Config.print_ident_value ident printer) ppf x;
  Format.fprintf ppf "@.@.";
;;

open Refinement_types;;
open Display_types;;
open Display_types_print;;
open Zonation;;
open Monitor;;
open Scalar_spec;;
open Components_spec;;
open Display_window;;

module Make =
  functor (Scal : SCALAR) ->
  functor (Comps : COMPONENTS) ->
struct
  module Dq = Display_quantity.Make (Scal) (Comps);;
  module Bb = Dq.Bb;;
  module St = Dq.St;;
  module H = Dq.H;;
  
  type state = St.t;;
  type history = H.t;;
  
  let print_iteration_report ppf i j =
    print_iteration_report Scal.print ppf (i, j);
  ;;
  
  (* Get configuration. *)
  
  let ppf = Config.get_formatter ();;
  
  let sz =
    (Translate.get_parameter_width (), Translate.get_parameter_height ())
  ;;
  print_report "image_size" print_image_size ppf sz;;
  
  let gm = Config.get_graphics_mode ();;
  let graphics =
    match gm with
    | No_graphics -> false
    | Parameter_graphics | Full_graphics -> true
  ;;
  
  let w_min_sz = Config.get_graphics_window_min_size ()
  and w_max_sz = Config.get_graphics_window_max_size ()
  ;;
  
  let draw set_and_draw q pps =
    set_and_draw q pps;
    Monitor.interruption ();
  ;;
  
  let init_quantity = draw Dq.set_quantity_and_draw;;
  
  let init sth =
    if graphics then
    begin
      Dq.setup gm w_min_sz w_max_sz sz;
      let { window_size = w_sz; scaling_factor = sf; _; } =
        get_graphics_elements () in
      print_report "window_size" print_window_size ppf w_sz;
      print_report "scaling_factor" print_scaling_factor ppf sf;
      if H.head_state_number sth > Nat._0_ then
        (* This means warm restart just occurred. *)
        let (zs, _j) = H.head_state sth in
        let ppps = Comps.map to_partition zs in
        init_quantity Param ppps;
        init_quantity Grad ppps;
    end;
  ;;
  
  let display_quantity = draw Dq.update_quantity_and_draw;;
  
  let display_param sth =
    let (zs, j) = H.head_state sth
    and i = H.head_state_number sth in
    print_iteration_report ppf i j;
    if graphics then
      let ppps = Comps.map to_partition zs in
      display_quantity Param ppps;
  ;;
  
  let display_grad (zs, _j) gs =
    if graphics then
      let ppps = Comps.map to_partition zs in
      let ppgs = Comps.map2 partition_of_gradient ppps gs in
      display_quantity Grad ppgs;
  ;;
end
;;
